<?php

declare(strict_types=1);

namespace BitBag\SyliusBonusPointsPlugin\EventListener;

use BitBag\SyliusBonusPointsPlugin\Calculator\DelegatingBonusPointsStrategyCalculatorInterface;
use BitBag\SyliusBonusPointsPlugin\Entity\BonusPointsInterface;
use BitBag\SyliusBonusPointsPlugin\Entity\BonusPointsStrategyInterface;
use BitBag\SyliusBonusPointsPlugin\Entity\CustomerBonusPointsInterface;
use BitBag\SyliusBonusPointsPlugin\Entity\CustomerBonusPointsStrategyInterface;
use BitBag\SyliusBonusPointsPlugin\Repository\BonusPointsStrategyRepositoryInterface;
use DateTime;
use Doctrine\ORM\EntityManagerInterface;
use Sylius\Component\Core\Model\CustomerInterface;
use Sylius\Component\Resource\Factory\FactoryInterface;
use Sylius\Component\Resource\Repository\RepositoryInterface;
use Symfony\Component\EventDispatcher\GenericEvent;

final class BonusPointsEventListener
{
    private $bonusPointsStrategyRepository;

    /** @var DelegatingBonusPointsStrategyCalculatorInterface */
    private $delegatingBonusPointsStrategyCalculator;

    /** @var FactoryInterface */
    private $bonusPointsFactory;

    /** @var EntityManagerInterface */
    private $bonusPointsEntityManager;

    /** @var RepositoryInterface */
    private $customerBonusPointsRepository;

    /** @var FactoryInterface */
    private $customerBonusPointsFactory;

    private $customerBonusPointsStrategyRepository;

    private $customerBonusPointsStrategyFactory;

    private $customerBonusPointsStrategyEntityManager;

    public function __construct(
        BonusPointsStrategyRepositoryInterface $bonusPointsStrategyRepository,
        DelegatingBonusPointsStrategyCalculatorInterface $delegatingBonusPointsStrategyCalculator,
        FactoryInterface $bonusPointsFactory,
        EntityManagerInterface $bonusPointsEntityManager,
        RepositoryInterface $customerBonusPointsRepository,
        FactoryInterface $customerBonusPointsFactory,
        RepositoryInterface $customerBonusPointsStrategyRepository,
        FactoryInterface $customerBonusPointsStrategyFactory,
        EntityManagerInterface $customerBonusPointsStrategyEntityManager
    ) {
        $this->bonusPointsStrategyRepository = $bonusPointsStrategyRepository;
        $this->delegatingBonusPointsStrategyCalculator = $delegatingBonusPointsStrategyCalculator;
        $this->bonusPointsFactory = $bonusPointsFactory;
        $this->bonusPointsEntityManager = $bonusPointsEntityManager;
        $this->customerBonusPointsRepository = $customerBonusPointsRepository;
        $this->customerBonusPointsFactory = $customerBonusPointsFactory;
        $this->customerBonusPointsStrategyRepository = $customerBonusPointsStrategyRepository;
        $this->customerBonusPointsStrategyEntityManager = $customerBonusPointsStrategyEntityManager;
        $this->customerBonusPointsStrategyFactory = $customerBonusPointsStrategyFactory;
    }

    public function __invoke(GenericEvent $event, string $eventName)
    {
        /** @var CustomerInterface $subject */
        $subject = $event->getSubject();

        /** @var BonusPointsStrategyInterface[] $bonusPointsStrategies */
        $bonusPointsStrategies = $this->bonusPointsStrategyRepository->findBy(['event' => $eventName]);

        $bonusPointsTotal = 0;
        foreach ($bonusPointsStrategies as $bonusPointsStrategy) {
            $bonusPointsTotal += $this->delegatingBonusPointsStrategyCalculator->calculate(
                $subject,
                $bonusPointsStrategy,
                0
            );
        }

        if ($bonusPointsTotal === 0) {
            return;
        }

        /** @var BonusPointsInterface $bonusPoints */
        $bonusPoints = $this->bonusPointsFactory->createNew();

        $bonusPoints->setPoints($bonusPointsTotal);
        $bonusPoints->setIsUsed(false);
        $bonusPoints->setExpiresAt((new DateTime())->modify('+1 year'));

        $this->getCustomerBonusPoints($subject)->addBonusPoints($bonusPoints);

        $this->bonusPointsEntityManager->persist($bonusPoints);
        $this->bonusPointsEntityManager->flush();

        foreach ($bonusPointsStrategies as $bonusPointsStrategy) {
            $this->addCustomerBonusPointsStrategy($subject, $bonusPointsStrategy);
        }
    }

    private function getCustomerBonusPoints(CustomerInterface $customer): CustomerBonusPointsInterface
    {
        /** @var CustomerBonusPointsInterface $customerBonusPoints */
        $customerBonusPoints = $this->customerBonusPointsRepository->findOneBy([
                'customer' => $customer,
            ]) ?? $this->customerBonusPointsFactory->createNew();

        if (null === $customerBonusPoints->getId()) {
            $customerBonusPoints->setCustomer($customer);

            $this->customerBonusPointsRepository->add($customerBonusPoints);
        }

        return $customerBonusPoints;
    }

    private function addCustomerBonusPointsStrategy(CustomerInterface $customer, BonusPointsStrategyInterface $bonusPointsStrategy): void
    {
        if (!($customerBonusPointsStrategy = $this->customerBonusPointsStrategyRepository->findOneBy([
            'customer' => $customer,
            'bonusPointsStrategy' => $bonusPointsStrategy,
        ]))) {
            /** @var CustomerBonusPointsStrategyInterface $customerBonusPointsStrategy */
            $customerBonusPointsStrategy = $this->customerBonusPointsStrategyFactory->createNew();
            $customerBonusPointsStrategy->setCustomer($customer);
            $customerBonusPointsStrategy->setBonusPointsStrategy($bonusPointsStrategy);

            $this->customerBonusPointsStrategyEntityManager->persist($customerBonusPointsStrategy);
            $this->customerBonusPointsStrategyEntityManager->flush();
        }
    }
}
